#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include "cdi.h"
#include "sedpc.h"
#include "cdiSedpcLib.h"
#define SD_WRITE 0x40
#define SD_READ  0x80

#define N_SEDAC_LINES 8
#define MAX_SEDAC_PUT 350   /* number of telegrams to put (4k FIFO 680 telegrams) */
#define WAIT_TIME_NS 1000000  // 1ms
int nSedacLines = N_SEDAC_LINES;
typedef struct
{
  int numberTelegrams, cdiLine;
	void *sedBuffer;
} SEDPC_DATA;
SEDPC_DATA sedacInfo[N_SEDAC_LINES] = { /* init */
  {0}, {0}, {0}, {0},  {0}, {0}, {0}, {0}
};

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodules","writemodules"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodules, writemodules, nCommands};

int sedac(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams);

int waitSedPCfinished()
{ int ln;
  CdiLineInfoBlk *busHandle;
  for (ln=0; ln<N_SEDAC_LINES; ln++)
  {
    if ( sedacInfo[ln].numberTelegrams <= 0 ) continue;
    if ((busHandle=cdiGetBusHandle(sedacInfo[ln].cdiLine)) == NULL) return -1;  /* something wrong!!  */
    cdiMutexLock(&(busHandle->lineMutex));
    cdiMutexUnlock(&(busHandle->lineMutex));
  }
  return 0;
}

__declspec(dllexport) int sedacScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  
  if (j == getcommands)
  {
    _snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }
  
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0);//atoi(c);

  if ( j == getlines )
  {
    _snprintf(result,resultBufferSize,"Sedac lines configured:\n");
    for (i=0, cnt=0; i<N_SEDAC_LINES; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       _snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > nSedacLines ) return no_such_line;
  if ( j == getcrates )
  {
    _snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
    for (crate=0, cnt=0; crate<32; crate++)
    {
      ret = sedac(line, crate, 253, CA_READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
      cnt++;
    }
    return 0;
  }
  else if ( j == getmodules )
  {
    crate = data[1];
    if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
    _snprintf(result,resultBufferSize,"Subaddresses on line %d crate %d:\n",line,crate);
    for (subaddr=0, cnt=0; subaddr<256; subaddr++)
    {
      ret = sedac(line, crate, subaddr, CA_READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
      cnt++;
    }
    return 0;
  }
  crate = data[1];
  subaddr = data[2];
  if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
  if ( j == readmodules )
  {
    _snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    ret = sedac(line, crate, subaddr, CA_READ, &rData, 1);
    if ((slen = strlen(result)) < resultBufferSize - 4) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
  }
  else if ( j == writemodules )
  {
    wData = data[3];
    if ( nData < 4 ) return argument_list_error;
    _snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    ret = sedac(line, crate, subaddr, CA_WRITE, &wData, 1);
    if ((slen = strlen(result)) < resultBufferSize - 4) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
  }  

#ifdef TESTCODE
  {
    _snprintf(result,resultBufferSize,"%d Sedac lines configured",nSedacLines);
    return 0;
  }
  strncpy(lclTarget,target,256);
  if ((c=strtok(lclTarget,":")) != NULL) 
  {
    line = atoi(c);
    if (line < 0 || line >= nSedacLines) return no_such_line;
    result[0] = 0;
    if ((c=strtok(NULL,":")) != NULL) 
    {
      crate = atoi(c);
      if ((c=strtok(NULL,":")) != NULL) 
      {
        subaddr = atoi(c);
        _snprintf(result,resultBufferSize,"addresses %d on line %d crate %d was found\n",subaddr,line,crate);
      }
      else
      {
        _snprintf(result,resultBufferSize,"sub-addresses configured on line %d crate %d:\n",line,crate);
        for (i=0; i<4; i++) // just write something out for testing ...
        {
          if ((slen = strlen(result)) > resultBufferSize - 8) break;
          _snprintf(&result[slen],resultBufferSize-slen,"%d ",i*16);
        }
        return 0;
      }
    }
    else
    {
      _snprintf(result,resultBufferSize,"crates configured on line %d :\n",line);
      for (i=0; i<8; i++) // just write something out for testing ...
      {
        if ((slen = strlen(result)) > resultBufferSize - 8) break;
        _snprintf(&result[slen],resultBufferSize-slen,"%d ",i*4);
      }
      return 0;
    }
  }
#endif
  
  return 0;
}
__declspec(dllexport) int filterSedacBus(int *parameter, int accessFlag)
{
  return 0;
}
__declspec(dllexport) int initSedacForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  cdi_mutex_t tmpMutex;
  int lastError, error = 0;
  /* try to get Mutex, is this a first application  */
  lastError = cdiMutexInit(&tmpMutex, cdiLine);
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);
  error = initSedac(busLine);
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
  if ( error == 0 ) sedacInfo[busLine-1].cdiLine = cdiLine;
  return error;
}

// if sync call, data not pass filter then repeat Transmit
__declspec(dllexport) int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= abs(pReq->timeout) )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}

__declspec(dllexport) void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  do
  {
    doSedacRequest(pReq, 1);
    if ( pReq->active == finishRequest ) return;

    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest ) return;
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  do
  {
    doSedacRequest(pReq, 2);
    if ( pReq->active == finishRequest ) return;

    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag && pReq->accessFlag != cdiWriteReadReadFlag)
  {
    pReq->active = finishRequest; /* finish request */
    return;
  }

  doSedacRequest(pReq, 3);
  pReq->active = finishRequest; /* finish request */
}

int clearSedac(short ln)
{ 
  short err=0;
  return err;
}
__declspec(dllexport) int initSedac(int ln)
{
  char *env;
  char str[80];
  short err;
  HANDLE hWnd = (HANDLE)_getpid( );

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  /* check environemnt variables to see what's connected: Stadmueller FIFO  */
  if ( ( env=getenv("SEDPC") ) == NULL )  return -non_existent;
  sprintf(str,"SEDACCDI:%d",ln);
  //SedIni(hWnd,str,&nlines,lines,devs,&err);
  //if ( err ) return -file_not_open;
  err = clearSedac((short)ln);
  /*  sedacInfo[ln - 1].numberTelegrams = 0;  /* set buffer size  */
  return err ? -not_ready : 0;
}
__declspec(dllexport) int exitSedac(int ln)
{
  return 0;
}

typedef struct
{
  WORD TypeAndCrate;
  WORD BaseSubAddress;
  WORD data;
} SedpcIn;
typedef struct
{
  WORD status;
  WORD data;
} SedpcOut;

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams)
{ 
  int i, steps, ret=0;
  static WORD storedWord[32][256];

  if (ntelegrams == 1)
  {
    if (access == CA_READ) *data = storedWord[cra][bsa]; //SedRd(lin,cra,bsa,(WORD *)data,(WORD *)&ret);
    else storedWord[cra][bsa] = *data;  //SedWr(lin,cra,bsa,*(WORD *)data,(WORD *)&ret);
//    Sleep(1);
    return ret;
  }

  if ( ntelegrams > sedacInfo[lin - 1].numberTelegrams && sedacInfo[lin - 1].numberTelegrams < MAX_SEDAC_PUT )  /* get buffer  */
  {
    if ( sedacInfo[lin - 1].numberTelegrams ) /* freee previous buffer */
    {
      free ( sedacInfo[lin - 1].sedBuffer );
      sedacInfo[lin - 1].numberTelegrams = 0;
      sedacInfo[lin - 1].sedBuffer = NULL;
    }
    steps = ntelegrams > MAX_SEDAC_PUT ? MAX_SEDAC_PUT : ntelegrams;
    if ( (sedacInfo[lin - 1].sedBuffer = (void *)malloc(steps*sizeof(SedpcIn)) ) == NULL ) return -out_of_local_memory;
    sedacInfo[lin - 1].numberTelegrams = steps;
  }
  steps = (ntelegrams - 1) / MAX_SEDAC_PUT + 1;
  for (i=0; i<ntelegrams; i++)
  {
    if (access == CA_READ) data[i] = storedWord[cra][bsa]; else storedWord[cra][bsa] = data[i];
  }
  Sleep(ntelegrams/4);
  return ret;
}

__declspec(dllexport) int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ret, length, writeOffset, done;
  void *pOut;

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag && pReq->accessFlag != cdiReadReadFlag) ? CA_WRITE : CA_READ;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? CA_WRITE : CA_READ;
  else
    access = pReq->accessFlag == cdiWriteReadReadFlag ? CA_READ : CA_WRITE;

  if ( access == CA_WRITE )
  {
    if ( pReq->accessFlag == cdiWriteWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag )
    {
      if ( step == 1 )
      {
        length = sizeof(short);
        pOut = (void *)pReq->pOutData;
        writeOffset = sedacSubOffsetSecond;
      }
      else
      {
        length = (pReq->accessFlag == cdiWriteReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : (pReq->outLength - sizeof(short));
        pOut = (void *)((char *)pReq->pOutData + sizeof(short));
        writeOffset = sedacSubOffsetThird;
      }
    }
    else
    {
      length = (pReq->accessFlag == cdiReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : pReq->outLength;
      pOut = (void *)pReq->pOutData;
      writeOffset = sedacSubOffsetSecond;
    }
  }
  else
  {
    done = 0;
    if ( pReq->accessFlag == cdiWriteReadReadFlag || pReq->accessFlag == cdiReadReadFlag )
    {
      if ( (pReq->accessFlag == cdiWriteReadReadFlag && step == 2) || (pReq->accessFlag == cdiReadReadFlag && step == 1 ) )
      {
        length = pReq->inLength < pReq->pDev->dataOuputSize ? pReq->inLength : pReq->pDev->dataOuputSize;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
        done = 1;
      }
      else if ( step == 3 )
      {
        if ( pReq->inLength <= pReq->pDev->dataOuputSize ) return pReq->errorCode;;
        length = pReq->inLength < (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize) ? (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize - pReq->inLength) : pReq->pDev->dataOuput2ndSize;
        pOut = (void *)((char *)pReq->pInData + pReq->pDev->dataOuputSize);
        writeOffset = sedacSubOffsetThird;
        done = 1;
      }
    }
    if ( done == 0 )
    {
        length = pReq->inLength;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
    }
  }

  //printf("try to get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);
  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  //printf("get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);

  ret = sedac((WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (WORD)pReq->pDev->cdiAddr[caAddrIndex],
//          (WORD)(pReq->pDev->cdiAddr[subAddrIndex] + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (WORD)(pReq->pDev->cdiAddr[subAddrIndex] + pReq->devParameters[writeOffset]), 
          (int)access,
          (WORD *)pOut,
          (int)length / sizeof(short) );

  if ( ret != 0 )
  {
    pReq->errorCode = ret < 0 ? -ret : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return pReq->errorCode;
}

