// gpioctl.h    Include file for Generic Port I/O Example Driver
//
// Define the IOCTL codes we will use.  The IOCTL code contains a command
// identifier, plus other information about the device, the type of access
// with which the file must have been opened, and the type of buffering.
//
// Robert B. Nelson (Microsoft)     March 1, 1993

// Device type           -- in the "User Defined" range."
#define GPD_TYPE 40000

// The IOCTL function codes from 0x800 to 0xFFF are for customer use.

#define IOCTL_GPD_READ_PORT_UCHAR \
    CTL_CODE( GPD_TYPE, 0x900, METHOD_BUFFERED, FILE_READ_ACCESS )

#define IOCTL_GPD_READ_PORT_USHORT \
    CTL_CODE( GPD_TYPE, 0x901, METHOD_BUFFERED, FILE_READ_ACCESS )

#define IOCTL_GPD_READ_PORT_ULONG \
    CTL_CODE( GPD_TYPE, 0x902, METHOD_BUFFERED, FILE_READ_ACCESS )

#define IOCTL_GPD_WRITE_PORT_UCHAR \
    CTL_CODE(GPD_TYPE,  0x910, METHOD_BUFFERED, FILE_WRITE_ACCESS)

#define IOCTL_GPD_WRITE_PORT_USHORT \
    CTL_CODE(GPD_TYPE,  0x911, METHOD_BUFFERED, FILE_WRITE_ACCESS)

#define IOCTL_GPD_WRITE_PORT_ULONG \
    CTL_CODE(GPD_TYPE,  0x912, METHOD_BUFFERED, FILE_WRITE_ACCESS)

//===> Sedac Device type           -- in the "User Defined" range."

// sedac error + data (ULONG for each telegram: SEDAC_DATA) transmissions
// or sedac block (SEDAC_PACKET) transmissions
#define IOCTL_SEDAC_DATA \
    CTL_CODE( GPD_TYPE, 0x920, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read how many telegrams in each queue
#define IOCTL_SEDAC_NUMBER_TELEGRAMS_IN_QUEUE \
    CTL_CODE( GPD_TYPE, 0x930, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read how many threads attached to this driver
#define IOCTL_SEDAC_NUMBER_THREADS \
    CTL_CODE( GPD_TYPE, 0x940, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read/write UseInterrupt flag
#define IOCTL_USE_INTERRUPT \
    CTL_CODE( GPD_TYPE, 0x950, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read/write PollingTimeUsec
#define IOCTL_POLLINGTIME_USEC \
    CTL_CODE( GPD_TYPE, 0x951, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read debug informatioj
#define IOCTL_SEDAC_DEBUG \
    CTL_CODE( GPD_TYPE, 0x952, METHOD_BUFFERED, FILE_ANY_ACCESS )

// read ip module ID information (output are all 12 bytes ID information)
#define IOCTL_SEDAC_IPID \
    CTL_CODE( GPD_TYPE, 0x953, METHOD_BUFFERED, FILE_ANY_ACCESS )

// sed3pci specific commands!!
// switch 32 or 8 Mhz mode
#define IOCTL_SEDAC_SWITCH32MHZ \
    CTL_CODE( GPD_TYPE, 0x960, METHOD_BUFFERED, FILE_ANY_ACCESS )
// write transmition sedac (data and address) array
#define IOCTL_SEDAC_SED3XMTARY \
    CTL_CODE( GPD_TYPE, 0x961, METHOD_BUFFERED, FILE_ANY_ACCESS )
// read receive sedac (data and status) array
#define IOCTL_SEDAC_SED3RCVARY \
    CTL_CODE( GPD_TYPE, 0x962, METHOD_BUFFERED, FILE_ANY_ACCESS )
// write transmition sedac 32 bits data(data and address) array
#define IOCTL_SEDAC_SED3XMT32 \
    CTL_CODE( GPD_TYPE, 0x963, METHOD_BUFFERED, FILE_ANY_ACCESS )
// read receive sedac 32 bits data (data and status) array
#define IOCTL_SEDAC_SED3RCV32 \
    CTL_CODE( GPD_TYPE, 0x964, METHOD_BUFFERED, FILE_ANY_ACCESS )

typedef struct  _GENPORT_WRITE_INPUT {
    ULONG   PortNumber;     // Port # to write to
    union   {               // Data to be output to port
        ULONG   LongData;
        USHORT  ShortData;
        UCHAR   CharData;
    };
} GENPORT_WRITE_INPUT;

// long input sedac data array
typedef struct  _SEDAC_ARY {
	USHORT  Data;		// input or output data
	USHORT	CommandReturn;	// (read, write or test) + cerateAdr and SubAdr (retuen with error)
} SEDAC_ARY, *PSEDAC_ARY;

// UCHAR mask bits definition for sedac command
#define MODE_READ	0x80						/* mode	*/
#define	MODE_WRITE	0x40
#define MODE_TEST	(MODE_READ + MODE_WRITE)
#define	MODE_CP		0x20
#define	MODE_R_CP	(MODE_CP + MODE_READ)
#define	MODE_W_CP	(MODE_CP + MODE_WRITE)
#define	MODE_TEST_CP	(MODE_CP + MODE_TEST)

#define	MASK_CRATE_ADR	0x1f
#define	MASK_COMMANDS	0xe0

typedef struct  _SEDAC_DATA_HEADER {
	USHORT	NumberOfError;	// return total number of error
	USHORT  SedacLine;		// line number start from 0....
	USHORT	numberRequested;
	USHORT	numberProcessed;
} SEDAC_DATA_HEADER, *PSEDAC_DATA_HEADER;

typedef struct  _SEDAC_DATA {
	USHORT  Data;		// input or output data
	USHORT	CommandReturn;	// (read, write or test) + cerateAdr and SubAdr (retuen with error)
	UCHAR	ExcessMode,	// input: (read, write or test.. commom pulse... ...)
			Line;		// line number
	UCHAR	SubAdr,		// SubAdr
			CaAdr;		// cerateAdr
} SEDAC_DATA, *PSEDAC_DATA;

#define SEDAC_DATA_SIZE			sizeof(SEDAC_DATA)
#define SEDAC_DATA_NUMBER_LONG	(sizeof(SEDAC_DATA)/sizeof(ULONG))
